/****************************************************************************/
/*                           the diy toolsuite                              */
/*                                                                          */
/* Jade Alglave, University College London, UK.                             */
/* Luc Maranget, INRIA Paris-Rocquencourt, France.                          */
/*                                                                          */
/* This C source is a product of litmus7 and includes source that is        */
/* governed by the CeCILL-B license.                                        */
/****************************************************************************/

/* Parameters */
#define OUT 1
#define SIZE_OF_TEST 5000
#define NUMBER_OF_RUN 200
#define AVAIL 24
#define N 2
#define NVARS 2
#define NEXE 12
#define NTHREADS 24
#define NOCCS 1
/* Includes */
#define KVM 1
#include <libcflat.h>
#include "kvm-headers.h"
#include "utils.h"

typedef uint32_t count_t;
#define PCTR PRIu32


/* Full memory barrier */
inline static void mbar(void) {
  asm __volatile__ ("dsb sy" ::: "memory");
}

/* Cache flush/fetch instructions */
#define CACHE_FLUSH 1
inline static void cache_flush(void *p) {
#ifdef CACHE_FLUSH
  asm __volatile__ ("dc civac,%[p]" :: [p] "r" (p) : "memory");
#endif
}


inline static void cache_touch(void *p) {
  asm __volatile__ ("prfm pldl1keep,[%[p]]" :: [p] "r" (p) : "memory");
}

inline static void cache_touch_store(void *p) {
  asm __volatile__ ("prfm pstl1keep,[%[p]]" :: [p] "r" (p) : "memory");
}



typedef struct {
  volatile int c,sense;
  int n ;
} sense_t;

static void barrier_init (sense_t *p,int n) {
  p->n = p->c = n;
  p->sense = 0;
}

/*
barrier_wait:
	ldr	w2, [x0,4]
.L28:
	ldaxr	w1, [x0]
	sub	w1, w1, #1
	stlxr	w3, w1, [x0]
	cbnz	w3, .L28
	cmp	w1, wzr
	ble	.L24
.L26:
	ldr	w1, [x0,4]
	cmp	w1, w2
	beq	.L26
	dmb	ish
.L24:
	ldr	w1, [x0,8]
	str	w1, [x0]
	mov	w1, 1
	dmb	ish
	sub	w2, w1, w2
	str	w2, [x0,4]
	dmb	ish
	ret
*/
static void barrier_wait(sense_t *p) {
  int sense = p->sense ;
  int r1,r3 ;
asm __volatile (
"dmb sy\n"
"0:\n\t"
"ldxr %w[r1],%[c]\n\t"
"sub %w[r1],%w[r1],#1\n\t"
"stxr %w[r3],%w[r1],%[c]\n\t"
"cbnz %w[r3],0b\n\t"
"cmp %w[r1],wzr\n\t"
"ble 2f\n\t"
"dmb sy\n"
"1:\n\t"
"ldr %w[r1],%[s]\n\t"
"cmp %w[r1],%w[ms]\n\t"
"beq 1b\n\t"
"b 3f\n"
"2:\n\t"
"dmb sy\n\t"
"ldr %w[r1],%[n]\n\t"
"str %w[r1],%[c]\n\t"
"mov %w[r3],#1\n\t"
"dmb sy\n\t"
"sub %w[ms],%w[r3],%w[ms]\n\t"
"str %w[ms],%[s]\n"
"3:\n\t"
"dsb sy\n\t"
: [r1] "=&r" (r1), [r3] "=&r" (r3)
: [c] "m" (p->c), [s] "m" (p->sense), [ms] "r" (sense), [n] "m" (p->n)
: "memory") ;
}
#if 0
/* C version */
static void barrier_wait(sense_t *p) {
  int sense = p->sense ;
  __sync_synchronize() ;
  int rem = __sync_add_and_fetch(&p->c,-1) ;
  if (rem > 0) {
    while (p->sense == sense) ;
    __sync_synchronize() ;
  } else {
    p->c = p->n ;
    __sync_synchronize() ;
    p->sense = 1-sense ;
    __sync_synchronize() ;
  }
}
#endif

/************/
/* Topology */
/************/

/*
 Topology: {{{0}, {1}, {2}, {3}, {4}, {5}, {6}, {7}, {8}, {9}, {10}, {11}, {12}, {13}, {14}, {15}, {16}, {17}, {18}, {19}, {20}, {21}, {22}, {23}}}
*/

static const int inst[] = {
// [[0],[1]]
5, 2, 4, 1, 9, 11, 0, 6, 6, 5, 7, 10, 8, 1, 3, 11, 10, 7, 4, 2, 8, 3, 9, 0,
};

static const int role[] = {
// [[0],[1]]
1, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 1, 1, 1, 1, 1, 0, 0, 1,
};

static const char *group[] = {
"[[0],[1]]",
};

#define SCANSZ 1
#define SCANLINE 24

/************/
/* Outcomes */
/************/

#define SOME_VARS 1
typedef struct {
  intmax_t *y,*x;
  pteval_t *pte_y,saved_pte_y,*pte_x,saved_pte_x;
} vars_t;


typedef struct {
  int out_1_x1;
  int out_0_x1;
  int y;
  int fault_P0_y,fault_P0_x,fault_P1_y;
} log_t;

static const int _idx_y = 1;
static const int _idx_pte_y = 2;
static const int _idx_x = 3;
static const int _idx_pte_x = 4;

static int idx_addr(intmax_t *v_addr,vars_t *p) {
  if (v_addr == NULL) { return 0;}
  else if (v_addr == p->y) return _idx_y;
  else if ((pteval_t *)v_addr == p->pte_y) return _idx_pte_y;
  else if (v_addr == p->x) return _idx_x;
  else if ((pteval_t *)v_addr == p->pte_x) return _idx_pte_x;
  else { fatal("???"); return -1;}
}

/* Dump of outcome */
static void pp_log(log_t *p) {
  printf("0:X1=%d;",p->out_0_x1);
  printf(" 1:X1=%d;",p->out_1_x1);
  printf(" y=%d;",p->y);
  printf(" %sfault(P0,y); %sfault(P0,x); %sfault(P1,y);",p->fault_P0_y?"":"~",p->fault_P0_x?"":"~",p->fault_P1_y?"":"~");
}

/* Equality of outcomes */
static int eq_log(log_t *p,log_t *q) {
  return
    p->out_0_x1 == q->out_0_x1 &&
    p->out_1_x1 == q->out_1_x1 &&
    p->y == q->y &&
    p->fault_P0_y == q->fault_P0_y &&
    p->fault_P0_x == q->fault_P0_x &&
    p->fault_P1_y == q->fault_P1_y &&
    1;
}

/* Fault Handling */
#define HAVE_FAULT_HANDLER 1

typedef struct { int instance,proc; } who_t;

static count_t nfaults[NTHREADS];
static who_t whoami[AVAIL];

typedef uint32_t ins_t; /* Type of instructions */

#define PRECISE 1
ins_t *label_ret[NTHREADS];

#define SEE_FAULTS 1


// Find index of some instruction in code, skipping 'skip' occurrences
static size_t find_ins(ins_t opcode,ins_t *p,int skip) {
  ins_t *q = p;

  for  ( ; *q != opcode || (skip-- > 0); q++);
  return q-p ;
}

static ins_t getnop(void) {
  ins_t *x1;
  ins_t r;
  asm __volatile__ (
  "adr %[x1],0f\n\t"
  "ldr %w[x2],[%[x1]]\n"
  "0:\n\t"
  "nop\n"
:[x1] "=&r" (x1),[x2] "=&r" (r)
:
: "cc","memory"
);
  return r;
}

typedef struct {
  int see_fault_P0_y,see_fault_P0_x,see_fault_P1_y;
} see_fault_t;

static see_fault_t *see_fault[NEXE];
static vars_t *vars_ptr[NEXE];

static void init_see_fault(see_fault_t *p) {
  p->see_fault_P0_y = 0;
  p->see_fault_P0_x = 0;
  p->see_fault_P1_y = 0;
}

static void record_fault(who_t *w,ins_t *pc,void *loc) {
  int i = w->instance;
  int idx_loc = idx_addr(loc,vars_ptr[i]);
  see_fault_t *sf = see_fault[i];
  switch (w->proc) {
  case 0: {
    if (idx_loc == _idx_y) {
      atomic_inc_fetch(&sf->see_fault_P0_y);
    }
    else if (idx_loc == _idx_x) {
      atomic_inc_fetch(&sf->see_fault_P0_x);
    }
    break;
  }
  case 1: {
    if (idx_loc == _idx_y) {
      atomic_inc_fetch(&sf->see_fault_P1_y);
    }
    break;
  }
  }
  atomic_inc_fetch(&nfaults[w->proc]);
}


static void fault_handler(struct pt_regs *regs,unsigned int esr) {
  struct thread_info *ti = current_thread_info();
  who_t *w = &whoami[ti->cpu];
  record_fault(w,read_elr_el1(),read_far());
#ifdef PRECISE
  regs->pc = (u64)label_ret[w->proc];
#endif
}

static void install_fault_handler(void) {
  install_exception_handler(EL1H_SYNC, ESR_EL1_EC_DABT_EL1,fault_handler);
}

static void pp_faults(void) {
  count_t total=0;
  for (int k=0 ; k < NTHREADS; k++) { total += nfaults[k]; }
  if (total > 0) {
    printf("Faults LB+dsb.sy+tlbi.ishppteoa %"PCTR"",total);
    for (int k = 0 ; k < NTHREADS ; k++) {
      count_t c = nfaults[k];
      if (c > 0) printf(" P%d:%"PCTR"",k,c);
    }
    printf("\n");
  }
}

inline static int final_cond(log_t *p) {
  return 1;
}

inline static int final_ok(int cond) {
  return !cond;
}


/**************/
/* Parameters */
/**************/

typedef enum { cignore, cflush, ctouch, cmax, } dir_t;

typedef struct {
  int part;
  int c_0_x,c_0_y,c_1_y;
} param_t;

static param_t param = {-1, -1, -1, -1,};

static int id(int x) { return x; }

static parse_param_t parse[] = {
  {"part",&param.part,id,SCANSZ},
  
  
  {"c_0_x",&param.c_0_x,id,cmax}, {"c_0_y",&param.c_0_y,id,cmax}, {"c_1_y",&param.c_1_y,id,cmax},
};

#define PARSESZ (sizeof(parse)/sizeof(parse[0]))

static void pp_param(param_t *p) {
  printf("{part=%d, c_0_x=%d, c_0_y=%d, c_1_y=%d}",p->part,p->c_0_x,p->c_0_y,p->c_1_y);
}

typedef struct {
  count_t groups[SCANSZ];
  count_t vars;
  count_t delays;
  count_t dirs[cmax][cmax][cmax];
} stats_t;

#define HASHSZ 432


/* Notice: this file contains public domain code by Bob Jenkins */

typedef struct {
  log_t key ;
  param_t p ;
  count_t c ;
  int ok ;
} entry_t ;

#ifdef KVM
static void pp_entry(entry_t *p, int verbose, const char **group) ;
#else
static void pp_entry(FILE *out,entry_t *p, int verbose, const char **group) ;
#endif

typedef struct {
  int nhash ;
  entry_t t[HASHSZ] ;
} hash_t ;

#ifdef KVM
static void pp_hash(hash_t *t,int verbose,const char **group) {
#else
static void pp_hash(FILE *fp,hash_t *t,int verbose,constchar **group) {
#endif
  for (int k = 0 ; k < HASHSZ ; k++) {
    entry_t *p = t->t+k ;
    if (p->c > 0) {
#ifdef KVM
      pp_entry(p,verbose,group) ;
#else
      pp_entry(fp,p,verbose,group) ;
#endif
    }
  }
}

#if 0
static void pp_hash_ok(FILE *fp,hash_t *t,char **group) {
  for (int k = 0 ; k < HASHSZ ; k++) {
    entry_t *p = t->t+k ;
    if (p->c > 0 && p->ok) pp_entry(fp,p,1,group) ;
  }
}
#endif


static void hash_init(hash_t *t) {
  t->nhash = 0 ;
  for (int k = 0 ; k < HASHSZ ; k++) t->t[k].c = 0 ;
}

/*
 Bob Jenkins hashword function, (Public domain)
 See <http://burtleburtle.net/bob/hash/doobs.html>
*/

#define rot(x,k) (((x)<<(k)) | ((x)>>(32-(k))))

#define mix(a,b,c) \
{ \
  a -= c;  a ^= rot(c, 4);  c += b; \
  b -= a;  b ^= rot(a, 6);  a += c; \
  c -= b;  c ^= rot(b, 8);  b += a; \
  a -= c;  a ^= rot(c,16);  c += b; \
  b -= a;  b ^= rot(a,19);  a += c; \
  c -= b;  c ^= rot(b, 4);  b += a; \
}

#define final(a,b,c) \
{ \
  c ^= b; c -= rot(b,14); \
  a ^= c; a -= rot(c,11); \
  b ^= a; b -= rot(a,25); \
  c ^= b; c -= rot(b,16); \
  a ^= c; a -= rot(c,4);  \
  b ^= a; b -= rot(a,14); \
  c ^= b; c -= rot(b,24); \
}

static uint32_t hashword(
const uint32_t *k,                   /* the key, an array of uint32_t values */
size_t          length)              /* the length of the key, in uint32_ts */
{
  uint32_t a,b,c;

  /* Set up the internal state */
  a = b = c = 0xdeadbeef ;

  /*------------------------------------------------- handle most of the key */
  while (length > 3)
  {
    a += k[0];
    b += k[1];
    c += k[2];
    mix(a,b,c);
    length -= 3;
    k += 3;
  }

  /*------------------------------------------- handle the last 3 uint32_t's */
  switch(length)                     /* all the case statements fall through */
  {
  case 3 : c+=k[2];
  case 2 : b+=k[1];
  case 1 : a+=k[0];
    final(a,b,c);
  case 0:     /* case 0: nothing left to add */
    break;
  }
  /*------------------------------------------------------ report the result */
  return c;
}

static uint32_t hash_log (log_t *key) {
  return hashword((uint32_t *)key,sizeof(log_t)/sizeof(uint32_t)) ;
}

static void hash_add(hash_t *t,log_t *key, param_t *v,count_t c,int ok) {
  uint32_t h = hash_log(key) ;
  h = h % HASHSZ ;
  for (int k = 0 ; k < HASHSZ ;  k++) {
    entry_t *p = t->t + h ;
    if (p->c == 0) { /* New entry */
      p->key = *key ;
      p->p = *v ;
      p->c = c ;
      p->ok = ok ;
      t->nhash++ ;
      return ;
    } else if (eq_log(key,&p->key)) {
      p->c += c ;
      return ;
    }
    h++ ;
    h %= HASHSZ ;
  }
#ifdef KVM
  printf("Hash table is full\n") ;
#else
  fprintf(stderr,"Hash table is full\n") ;
#endif
  exit(2) ;
}

static void hash_adds(hash_t *t, hash_t *f) {
  for (int k = 0 ; k < HASHSZ ; k++) {
    entry_t *p = f->t+k ;
    if (p->c > 0) {
      hash_add(t,&p->key,&p->p,p->c,p->ok) ;
    }
  }
}

static void pp_entry(entry_t *p, int verbose, const char **group) {
  printf("%-6" PCTR "%c>",p->c,p->ok ? '*' : ':');
  pp_log(&p->key);
  if (verbose) {
    printf(" # ");
    pp_param(&p->p);
    printf(" %s",group[p->p.part]);
  }
  printf("%c",'\n');
}

static void set_feature(int _role) { }

/***************/
/* Memory size */
/***************/

/* Page size line */
#define LINE LITMUS_PAGE_SIZE

static void vars_init(vars_t *_vars,intmax_t *_mem) {
  const size_t _sz = LINE/sizeof(intmax_t);
  pteval_t *_p;

  _vars->y = _mem;
  _vars->pte_y = _p = litmus_tr_pte((void *)_mem);
  _vars->saved_pte_y = *_p;
  _mem += _sz ;
  _vars->x = _mem;
  _vars->pte_x = _p = litmus_tr_pte((void *)_mem);
  _vars->saved_pte_x = *_p;
  _mem += _sz ;
}


typedef struct {
  int id ;
  intmax_t *mem;
  log_t out;
#ifdef SOME_PTR
  log_ptr_t out_ptr;
#endif
#ifdef SOME_VARS
  vars_t v;
#endif
#ifdef SEE_FAULTS
  see_fault_t f;
#endif
#ifdef HAVE_TIMEBASE
  tb_t next_tb;
#endif
  hash_t t;
  sense_t b;
  param_t p; // For random scan
} ctx_t ;


static void instance_init (ctx_t *p, int id, intmax_t *mem) {
  p->id = id ;
  p->mem = mem ;
  hash_init(&p->t) ;
  barrier_init(&p->b,N) ;
#ifdef KVM
#ifdef SOME_VARS
  vars_init(&p->v,mem);
#endif
#endif
}

/******************/
/* Global context */
/******************/

#define LINESZ (LINE/sizeof(intmax_t))
#define MEMSZ ((NVARS*NEXE+1)*LINESZ)

static intmax_t mem[MEMSZ] ;

typedef struct global_t {
  /* Command-line parameter */
  param_t *param ;
  parse_param_t *parse ;
#ifdef KVM
  int volatile over;
#endif
  /* Topology */
  const int *inst, *role ;
  const char **group ;
  /* memory */
  intmax_t *mem ;
  /* Cache control */
#ifdef ACTIVE
  active_t *active;
#endif
  /* Runtime control */
  int verbose ;
  int size,nruns,nexe,noccs ;
  int do_scan ;
  /* Synchronisation for all threads */
  volatile int go ; /* First synchronisation */
  sense_t gb ;    /* All following synchronisation */
  /* Count 'interesting' outcomes */
  volatile int ok ;
  /* Times for timeout */
  tsc_t start,now ;
  /* All instance contexts */
  ctx_t ctx[NEXE] ; /* All test instance contexts */
  hash_t hash ;     /* Sum of outcomes */
  /* statistics */
  stats_t stats ;
} global_t ;

static global_t global  =
  { &param, &parse[0],
#ifdef KVM
    0,
#endif
    inst, role, group, mem,
#ifdef ACTIVE
    active,
#endif
  };

static void init_global(global_t *g,int id) {
  if (id == 0) {
#ifdef TIMELIMIT
    /* Starting time */
    g->start = timeofday() ;
#endif
    /* Global barrier */
    barrier_init(&g->gb,AVAIL) ;
    /* Align  to cache line */
    uintptr_t x = (uintptr_t)(g->mem) ;
    x += LINE-1 ; x /=  LINE ; x *= LINE ;
    intmax_t *m = (intmax_t *)x ;
    /* Instance contexts */
    for (int k = 0 ; k < NEXE ; k++) {
      instance_init(&g->ctx[k],k,m) ;
      m += NVARS*LINESZ ;
    }
    mbar() ;
    g->go = 1 ;
  } else {
    while (g->go == 0) ;
    mbar() ;
  }
}

/******************/
/* Thread context */
/******************/

typedef struct {
  int id ;
  st_t seed ;
  int role,inst ;
  ctx_t *ctx ;
#ifdef ACTIVE
  active_t *act;
#endif
} thread_ctx_t ;


static void set_role(global_t *g,thread_ctx_t *c,int part) {
  barrier_wait(&g->gb) ;
  int idx = SCANLINE*part+c->id ;
  int inst = g->inst[idx] ;
  if (0 <= inst && inst < g->nexe) {
    c->inst = inst ;
    c->ctx = &g->ctx[inst] ;
    c->role = g->role[idx] ;
#ifdef KVM
    set_feature(c->role) ;
#ifdef HAVE_FAULT_HANDLER
    whoami[c->id].instance = inst ;
    whoami[c->id].proc = c->role ;
#ifdef SEE_FAULTS
    if (c->role == 0) {
      vars_ptr[inst] = &c->ctx->v;
      see_fault[inst] = &c->ctx->f;
    }
#endif
#endif
#endif
#ifdef ACTIVE
    c->act = &g->active[part] ;
#endif
  } else {
    c->ctx = NULL ;
    c->role = -1 ;
  }
  barrier_wait(&g->gb) ;
}

/*************/
/* Test code */
/*************/

typedef void (*code0_t)(int *x,int *y,int *out_0_x1);

__attribute__ ((noinline))static void code0(int *x,int *y,int *out_0_x1) {
  int trashed_x2;
asm __volatile__ (
"\n"
"#START _litmus_P0\n"
"#_litmus_P0_0\n\t"
"nop\n"
"#_litmus_P0_1\n\t"
"ldr %w[x1],[%[x0]]\n"
"#_litmus_P0_2\n\t"
"dsb sy\n"
"#_litmus_P0_3\n\t"
"mov %w[x2],#1\n"
"#_litmus_P0_4\n\t"
"str %w[x2],[%[x3]]\n"
"#_litmus_P0_5\n\t"
"nop\n"
"#END _litmus_P0\n\t"
:[x1] "=&r" (*out_0_x1),[x2] "=&r" (trashed_x2)
:[x0] "r" (x),"[x1]" (0),[x3] "r" (y)
:"cc","memory"
);
}

typedef void (*code1_t)(int *y,pteval_t *pte_x,pteval_t _pteval0,int *out_1_x1);

__attribute__ ((noinline))static void code1(int *y,pteval_t *pte_x,pteval_t _pteval0,int *out_1_x1) {
  void* trashed_x4;
asm __volatile__ (
"\n"
"#START _litmus_P1\n"
"#_litmus_P1_0\n\t"
"nop\n"
"#_litmus_P1_1\n\t"
"ldr %w[x1],[%[x0]]\n"
"#_litmus_P1_2\n\t"
"lsr %[x4],%[x0],#12\n"
"#_litmus_P1_3\n\t"
"dsb ish\n"
"#_litmus_P1_4\n\t"
"tlbi vaae1is,%[x4]\n"
"#_litmus_P1_5\n\t"
"dsb ish\n"
"#_litmus_P1_6\n\t"
"str %[x3],[%[x2]]\n"
"#_litmus_P1_7\n\t"
"nop\n"
"#END _litmus_P1\n\t"
:[x1] "=&r" (*out_1_x1),[x4] "=&r" (trashed_x4)
:[x0] "r" (y),"[x1]" (0),[x2] "r" (pte_x),[x3] "r" (_pteval0)
:"cc","memory"
);
}

inline static int do_run(thread_ctx_t *_c, param_t *_p,global_t *_g) {
  int _ok = 0;
  int _role = _c->role;
  if (_role < 0) return _ok;
  ctx_t *_ctx = _c->ctx;
  sense_t *_b = &_ctx->b;
  log_t *_log = &_ctx->out;
  vars_t *_vars = &_ctx->v;
  int *y = (int *)_vars->y;
  int *x = (int *)_vars->x;
  barrier_wait(_b);
  switch (_role) {
  case 0: {
    init_see_fault(&_ctx->f);
    *x = 0;
    litmus_flush_tlb((void *)x);
    barrier_wait(_b);
    if (_p->c_0_x == ctouch) cache_touch((void *)x);
    else if (_p->c_0_x == cflush) cache_flush((void *)x);
    if (_p->c_0_y == ctouch) cache_touch((void *)y);
    else if (_p->c_0_y == cflush) cache_flush((void *)y);
    barrier_wait(_b);
    code0(x,y,&_log->out_0_x1);
    _log->fault_P0_y = _ctx->f.see_fault_P0_y?1:0;
    _log->fault_P0_x = _ctx->f.see_fault_P0_x?1:0;
    barrier_wait(_b);
    *(_vars->pte_x) = _vars->saved_pte_x;
    litmus_flush_tlb((void *)x);
    barrier_wait(_b);
    int _cond = final_ok(final_cond(_log));
    hash_add(&_ctx->t,_log,_p,1,_cond);
    if (_cond) {
      _ok = 1;
      (void)__sync_add_and_fetch(&_g->stats.groups[_p->part],1);
      (void)__sync_add_and_fetch(&_g->stats.dirs[_p->c_0_x][_p->c_0_y][_p->c_1_y],1);
    }
    break; }
  case 1: {
    *y = 0;
    litmus_flush_tlb((void *)y);
    barrier_wait(_b);
    if (_p->c_1_y == ctouch) cache_touch((void *)y);
    else if (_p->c_1_y == cflush) cache_flush((void *)y);
    barrier_wait(_b);
    code1(y,_vars->pte_x,_vars->saved_pte_y,&_log->out_1_x1);
    _log->fault_P1_y = _ctx->f.see_fault_P1_y?1:0;
    barrier_wait(_b);
    *(_vars->pte_y) = _vars->saved_pte_y;
    litmus_flush_tlb((void *)y);
    _log->y = *y;
    barrier_wait(_b);
    break; }
  }
  return _ok;
}

static void init_labels(void) {
  ins_t nop = getnop();
  label_ret[0] = ((ins_t *)code0)+find_ins(nop,(ins_t *)code0,1);
  label_ret[1] = ((ins_t *)code1)+find_ins(nop,(ins_t *)code1,1);
}

/*******************/
/* Forked function */
/*******************/

inline static int comp_param (st_t *seed,int *g,int max,int delta) {
  int tmp = *g;
  return tmp >= 0 ? tmp : delta+rand_k(seed,max-delta);
}

static void choose_params(global_t *g,thread_ctx_t *c,int part) {
  int _role = c->role;
  if (_role < 0) return;
  ctx_t *ctx = c->ctx;

  for (int _s=0 ; _s < g->size ; _s++) {
    barrier_wait(&ctx->b);
    switch (_role) {
    case 0:
      ctx->p.c_0_x = comp_param(&c->seed,&g->param->c_0_x,cmax,1);
      ctx->p.c_1_y = comp_param(&c->seed,&g->param->c_1_y,cmax,1);
      break;
    case 1:
      ctx->p.part = part;
      ctx->p.c_0_y = comp_param(&c->seed,&g->param->c_0_y,cmax,1);
      break;
    }
    (void)do_run(c,&ctx->p,g);
  }
}

static void choose(int id,global_t *g) {
  param_t *q = g->param;
  thread_ctx_t c; c.id = c.seed = id;

  if (q->part >=0) {
    set_role(g,&c,q->part);
    for (int nrun = 0; nrun < g->nruns ; nrun++) {
      choose_params(g,&c,q->part);
    }
  } else {
    st_t seed = 0;
    for (int nrun = 0; nrun < g->nruns ; nrun++) {
      int part = rand_k(&seed,SCANSZ);
      set_role(g,&c,part);
      choose_params(g,&c,part);
    }
  }
}

typedef struct {
  int id;
  global_t *g;
} zyva_t;

static void zyva(void *_a) {
  zyva_t *a = (zyva_t*)_a;
  int id = a->id;
  global_t *g = a->g;
  set_hahd_bits(0b00);
  install_fault_handler();
  init_global(g,id);
  choose(id,g);
  mbar();
  atomic_inc_fetch(&g->over);
}

static void feature_check(void) {
}

#define RUN run
#define MAIN 1

#define ENOUGH 10

static void postlude(global_t *g,count_t p_true,count_t p_false,tsc_t total) {
  hash_t *hash = &g->hash ;
  printf("Test LB+dsb.sy+tlbi.ishppteoa Required\n");
  printf("Histogram (%d states)\n",hash->nhash);
  pp_hash(hash,g->verbose > 1,g->group);
  int cond = p_true == 0;
  printf("%s\n",cond?"Ok":"No");
  printf("\nWitnesses\n");
  printf("Positive: %" PCTR ", Negative: %" PCTR "\n",p_false,p_true);
  printf("Condition %s is %svalidated\n","forall (true)",cond ? "" : "NOT ");
  printf("Hash=53bf12092eea6260bd65f65357a6ad0a\n");
  printf("Variant=precise\n");
  printf("Cycle=Rfe TLBI.ISHdRWPPteOA RfePteOAP DSB.SYdRW\n");
  printf("Generator=diycross7 (version 7.56+02~dev)\n");
  printf("Com=Rf Rf\n");
  printf("Orig=DSB.SYdRW Rfe TLBI.ISHdRWPPteOA RfePteOAP\n");
  count_t cond_true = p_false;
  count_t cond_false = p_true;
  printf("Observation LB+dsb.sy+tlbi.ishppteoa %s %" PCTR " %" PCTR "\n",!cond_true ? "Never" : !cond_false ? "Always" : "Sometimes",cond_true,cond_false);
  pp_faults();
  puts("Time LB+dsb.sy+tlbi.ishppteoa ");
  emit_double(tsc_millions(total));
  puts("\n");
}



/***************/
/* Entry point */
/***************/
#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS 0
#endif

static zyva_t arg[AVAIL];
#ifndef KVM
static pthread_t th[AVAIL];
#endif

#ifdef KVM
int RUN(int argc,char **argv) ;
int RUN(int argc,char **argv) {
  feature_check();
#else
int RUN(int argc,char **argv,FILE *out) ;
int RUN(int argc,char **argv,FILE *out) {
#endif
#ifdef OUT
  opt_t def = { 0, NUMBER_OF_RUN, SIZE_OF_TEST, AVAIL, NEXE, mode_random, };
  opt_t d = def ;
  char *prog = argv[0] ;
  char **p = parse_opt(argc,argv,&def,&d) ;
  int n_exe = d.n_exe ;
  if (d.avail != AVAIL) n_exe = d.avail / N ;
  if (n_exe < 1) n_exe = 1 ;
  global.verbose = d.verbose;
  global.nexe = n_exe;
  global.noccs = NOCCS ;
  global.nruns = d.max_run;
  global.size = d.size_of_test;
  global.do_scan = d.mode == mode_scan ;
  if (global.verbose) {
#ifdef KVM
    printf("%s: n=%d, r=%d, s=%d, %s\n",prog,global.nexe,global.nruns,global.size,global.do_scan ? "+sp" : "+rp");
#else
    fprintf(stderr,"%s: n=%i, r=%i, s=%i, %s\n",prog,global.nexe,global.nruns,global.size,global.do_scan ? "+sp" : "+rp");
#endif
  }
  parse_param(prog,global.parse,PARSESZ,p) ;
#ifdef PRELUDE
#ifndef KVM
  prelude(out) ;
#endif
#endif
  tsc_t start = timeofday();
#else
  global.verbose = 0 ;
  global.nexe = NEXE ;
  global.noccs = NOCCS ;
  global.nruns = NUMBER_OF_RUN ;
  global.size = SIZE_OF_TEST ;
  global.do_scan = 0;
#endif
  for (int id=0; id < AVAIL ; id++) {
    arg[id].id = id;
    arg[id].g = &global;
  }
#ifdef KVM
  /* "spawn" downwards as id 0 is not asynchornous */
  init_labels();
  global.over = 0 ;
  for (int id = AVAIL-1 ; id >= 0 ; id--) on_cpu_async(id,zyva,&arg[id]);
  while (global.over < AVAIL) mdelay(500);
#else
  for (int id=0; id < AVAIL ; id++) launch(&th[id],zyva,&arg[id]);
  for (int id=0; id < AVAIL ; id++) join(&th[id]);
#endif
  int nexe = global.nexe ;
  hash_init(&global.hash) ;
  for (int k=0 ; k < nexe ; k++) {
    hash_adds(&global.hash,&global.ctx[k].t) ;
  }
#ifdef OUT
  tsc_t total = timeofday()-start;
  count_t p_true = 0, p_false = 0;
  for (int k = 0 ; k < HASHSZ ; k++) {
    entry_t *e = &global.hash.t[k];
    if (e->ok) {
      p_true += e->c ;
    } else {
      p_false += e->c;
    }
  }
#ifdef KVM
  postlude(&global,p_true,p_false,total);
#else
  postlude(out,&global,p_true,p_false,total);
#endif
#endif
  return EXIT_SUCCESS;
}

#ifdef MAIN
int main (int argc,char **argv) {
#ifdef KVM
  litmus_init();
  return RUN(argc,argv) ;
#else
  return RUN(argc,argv,stdout) ;
#endif
}
#endif
